

extern SV *qid;
extern SV *query;

void boot_query(void);

void savestate_query(void);

void closequery(void);

void cutquery(void);

void testnoquery(void);

void testquery(void);

int isquery(void);
